--[[								-*- Mode: lua -*-
Baggins_ItemRack

Author: Daphne Pfister <tiiroib02@sneakemail.com>

Thanks to Nargiddley for writing Baggins and Gello for writing ItemRack

License

Copyright (c) 2007 Daphne Pfister

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
]]

Baggins_ItemRack = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0","AceHook-2.1")

local Baggins_ItemRack = Baggins_ItemRack

local dewdrop = AceLibrary:HasInstance("Dewdrop-2.0") and AceLibrary("Dewdrop-2.0")
local L = AceLibrary("AceLocale-2.2"):new("Baggins_ItemRack")
local sieve
local item_rack_character
local other_characters_scanned
local items_table = {}
local sets_table = {}
local character_table = {}
local old_version

L:RegisterTranslations("enUs", function()
	return {
		["ItemRack Character:%s Set:%s"] = true,
		["ItemRack Set:%s"] = true,
		["Unknown"] = true,
		["ItemRack Sets"] = true,
		["Character Name"] = true,
		["Set Name"] = true,
		["Any"] = true,
		["None"] = true,
		["Realm: %s"] = true,
		["ItemRack"] = true,
		["Filter by ItemRack set"] = true,
		["Fuzzy match"] = true,
		["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = true,
	};
end)

-- You have my permission to fix any of these locales without
-- asking me and check them into the SVN.
L:RegisterTranslations("esES", function()
	return {
		--["ItemRack Character:%s Set:%s"] = true,
		--["ItemRack Set:%s"] = true,
		--["Unknown"] = true,
		--["ItemRack Sets"] = true,
		--["Character Name"] = true,
		--["Set Name"] = true,
		--["Any"] = true,
		["None"] = "Ninguno", -- check
		--["Realm: %s"] = true,
		--["ItemRack"] = true,
		--["Filter by ItemRack set"] = true,
		--["Fuzzy match"] = true,
		--["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = true,
	};
end)

L:RegisterTranslations("deDE", function()
	return {
		--["ItemRack Character:%s Set:%s"] = true,
		--["ItemRack Set:%s"] = true,
		["Unknown"] = "Unbekannt", -- check
		--["ItemRack Sets"] = true,
		--["Character Name"] = true,
		--["Set Name"] = true,
		--["Any"] = true,
		["None"] = "Keine",
		--["Realm: %s"] = true,
		--["ItemRack"] = true,
		--["Filter by ItemRack set"] = true,
		--["Fuzzy match"] = true,
		--["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = true,
	};
end)

L:RegisterTranslations("frFR", function()
	return {
		--["ItemRack Character:%s Set:%s"] = true,
		--["ItemRack Set:%s"] = true,
		["Unknown"] = "Inconnu", -- check
		--["ItemRack Sets"] = true,
		--["Character Name"] = true,
		--["Set Name"] = true,
		--["Any"] = true,
		["None"] = "Aucun", -- check
		--["Realm: %s"] = true,
		--["ItemRack"] = true,
		--["Filter by ItemRack set"] = true,
		--["Fuzzy match"] = true,
		--["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = true,
	};
end)

L:RegisterTranslations("koKR", function()
	return {
		--["ItemRack Character:%s Set:%s"] = true,
		--["ItemRack Set:%s"] = true,
		["Unknown"] = "가격미확인", -- check
		--["ItemRack Sets"] = true,
		--["Character Name"] = true,
		--["Set Name"] = true,
		--["Any"] = true,
		["None"] = "없음", -- check
		--["Realm: %s"] = true,
		--["ItemRack"] = true,
		--["Filter by ItemRack set"] = true,
		--["Fuzzy match"] = true,
		--["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = true,
	};
end)

L:RegisterTranslations("znCN", function()
	return {
		--["ItemRack Character:%s Set:%s"] = true,
		--["ItemRack Set:%s"] = true,
		["Unknown"] = "未知", -- check
		--["ItemRack Sets"] = true,
		--["Character Name"] = true,
		--["Set Name"] = true,
		--["Any"] = true,
		["None"] = "无", -- check
		--["Realm: %s"] = true,
		--["ItemRack"] = true,
		--["Filter by ItemRack set"] = true,
		--["Fuzzy match"] = true,
		--["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = true,
	};
end)

L:RegisterTranslations("znTW", function()
	return {
		--["ItemRack Character:%s Set:%s"] = true,
		--["ItemRack Set:%s"] = true,
		["Unknown"] = "未知", -- check
		--["ItemRack Sets"] = true,
		--["Character Name"] = true,
		--["Set Name"] = true,
		--["Any"] = true,
		["None"] = "無", -- check
		--["Realm: %s"] = true,
		--["ItemRack"] = true,
		--["Filter by ItemRack set"] = true,
		--["Fuzzy match"] = true,
		--["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = true,
	};
end)

local function linkToId(link)
	if link then
		if old_version then
			return link:match("item:([%-%d]+:[%-%d]+:[%-%d]+):[%-%d]+")
		else
			return link:match("item:([%-%d:]+):%-?%d+")
		end
	end
	return nil
end

local function idToFuzzy(id)
	if type (id) == "string" then
		return id:match("(%-?%d+)")
	else
		return id
	end
end

local function Matches(bag,slot,rule)
	local id = linkToId(GetContainerItemLink(bag, slot))
	if not id then return end

	local set = rule.set
	if not set then return end

	local invert = nil
	if set == "None" then
		set = "Any"
		invert = true
	end
	if rule.fuzzy then
		id = idToFuzzy(id)
	end

	local res = items_table
	if res and old_version then
		local character = rule.user
		if not character then return end
		if character == "Self" then character = item_rack_character end
		if not character then return end
		res = res[character]
	end
	res = res and res[id]
	if set ~= "Any" then res = res and res[set] end
	if invert then res = not res end
	return res
end

local function GetName(rule)
	local msg
	if old_version then
		msg = L["ItemRack Character:%s Set:%s"]:format(rule.user or L["Unknown"], rule.set or L["Unknown"])
	else
		msg = L["ItemRack Set:%s"]:format(rule.set or L["Unknown"])
	end
	if rule.fuzzy then
		msg = msg.." "..L["Fuzzy match"]
	end
	return msg
end

local function CleanRule(rule)
	if old_version then
		rule.user = "Self"
	end
	rule.set = "Any"
	sieve:OnRuleChanged()
end

local function setRuleUser(rule, user)
	rule.user = user
	rule.set = "Any"
	sieve:OnRuleChanged()
end
	
local function setRuleSet(rule, set)
	rule.set = set
	sieve:OnRuleChanged()
end

local function toggleRuleFuzzy(rule)
	rule.fuzzy = not rule.fuzzy
	sieve:OnRuleChanged()
end

local function DewDropOptions(rule, level, value)
	if old_version then
		if level == 1 then
			dewdrop:AddLine('text', L["ItemRack Sets"], 'isTitle', true)
			dewdrop:AddLine('text', L["Character Name"], 'hasArrow', true,'value',"ItemRackCharacter")
			dewdrop:AddLine('text', L["Set Name"], 'hasArrow', true, 'value', "ItemRackSet")
			dewdrop:AddLine('text', L["Fuzzy match"], 'checked', rule.fuzzy, 'func',toggleRuleFuzzy,'arg1',rule,
					'tooltipTitle', L["Fuzzy match"],
					'tooltipText',L["Match by the item number, ignoring enchantments, variations, armor kits, etc."])
		elseif level == 2 and value == "ItemRackCharacter" then
			dewdrop:AddLine('text', L["Any"], "checked", rule.user == "Any", 'func',setRuleUser,'arg1',rule,'arg2',"Any")
			dewdrop:AddLine('text', L["Self"], "checked", rule.user == "Self", 'func',setRuleUser,'arg1',rule,'arg2',"Self")
			local realm
			for realm in pairs(character_table) do
				dewdrop:AddLine('text', L["Realm: %s"]:format(realm), 'hasArrow', true, 'value', 'Realm:'..realm)
			end
		elseif level == 3 then
			local _,_,realm = value:find("Realm:(.*)")
			if realm then
				local character
				local full
				for character,full in pairs(character_table[realm]) do
					dewdrop:AddLine('text', character, "checked", rule.user == full, 'func',setRuleUser,'arg1',rule,'arg2',full)
				end
			end
		elseif level == 2 and value == "ItemRackSet" then
			dewdrop:AddLine('text', L["Any"], "checked", rule.set == "Any", 'func',setRuleSet,'arg1',rule,'arg2',"Any")
			dewdrop:AddLine('text', L["None"], "checked", rule.set == "None", 'func',setRuleSet,'arg1',rule,'arg2',"None")
			local character = rule.user
			-- Treat self as "Any" here.. so that we include sets from other characters...
			-- since this is probably a shared category.
			if character and character == "Self" then character = "Any" end
			local charactersets = sets_table[character]
			if charactersets then
				local setnames = {}
				for set in pairs(charactersets) do
					table.insert(setnames,set)
				end
				table.sort(setnames)
				for _,set in ipairs(setnames) do
					dewdrop:AddLine('text', set, "checked", rule.set == set, 'func',setRuleSet,'arg1',rule,'arg2',set)
				end
			end
		end
	elseif level == 1 then
		dewdrop:AddLine('text', L["ItemRack Sets"], 'isTitle', true)
		dewdrop:AddLine("text", L["Any"], "checked", rule.set == "Any", 'func',setRuleSet,'arg1',rule,'arg2',"Any")
		dewdrop:AddLine("text", L["None"], "checked", rule.set == "None", 'func',setRuleSet,'arg1',rule,'arg2',"None")
		local setnames = {}
		for set in pairs(sets_table) do
			table.insert(setnames,set)
		end
		table.sort(setnames)
		for _,set in ipairs(setnames) do
			dewdrop:AddLine('text', set, "checked", rule.set == set, 'func',setRuleSet,'arg1',rule,'arg2',set)
		end
		dewdrop:AddLine('text', "", 'isTitle', true)
		dewdrop:AddLine('text', L["Fuzzy match"], 'checked', rule.fuzzy, 'func',toggleRuleFuzzy,'arg1',rule,
				'tooltipTitle', L["Fuzzy match"],
				'tooltipText',L["Match by the item number, ignoring enchantments, variations, armor kits, etc."])
	end
end

function Baggins_ItemRack:OnEnable()
	sieve = AceLibrary("Sieve-0.1")

	item_rack_character = UnitName("player").." of "..GetRealmName()

	if ItemRack_Version and ItemRack_Version < 2 then
		old_version = true
	end

	self:CacheSets()

	if old_version then
		self:Hook("ItemRack_Sets_Initialize")
		self:Hook("ItemRack_Sets_Save_OnClick")
		self:Hook("ItemRack_Sets_Remove_OnClick")
	else
		self:Hook(ItemRack,"InitCore")
		if ItemRackOpt then
			self:HookOptions()
		else
			self:Hook(ItemRack,"ToggleOptions")
		end
	end

	sieve:AddCustomRule("ItemRack", {
		DisplayName = L["ItemRack"],
		Description = L["Filter by ItemRack set"],
		Matches = Matches,
		GetName = GetName,
		CleanRule = CleanRule,
		DewDropOptions = DewDropOptions
	})
end

function Baggins_ItemRack:CacheSetsForCharacter(u)
	local i,setname,set

	--self:Print("Caching: "..u)
	local _,_,character,realm = u:find("(.*) of (.*)")
	if not character or not realm then
		-- Unexpected format for ItemRack character
		character = u
		realm = "Other"
	end
	if not character_table[realm] then
		character_table[realm] = {}
	end
	character_table[realm][character] = u

	-- Remove any items mappings for this character
	local item_table = {}
	items_table[u] = item_table
	local set_table = {}
	sets_table[u] = set_table

	local charactersets = Rack and Rack_User and Rack_User[u] and Rack_User[u].Sets
	if not charactersets then return end

	local other = (u ~= item_rack_character)
	local other_item_table
	other_item_table = items_table.Other
	if not other_item_table then
		other_item_table = {}
		items_table.Other = other_item_table
	end
	local other_set_table
	other_set_table = sets_table.Other
	if not other_set_table then
		other_set_table = {}
		sets_table.Other = other_set_table
	end

	for setname, set in pairs(charactersets) do
		if not setname:find("^ItemRack-") and not setname:find("^Rack-") then
			set_table[setname] = true
			if other then
				other_set_table[setname] = true
			end
			for i=0,19 do
				local setitem = set[i] and set[i].id
				if setitem then
					if not item_table[setitem] then item_table[setitem] = {} end
					item_table[setitem][setname] = true
					if other then
						if not other_item_table[setitem] then other_item_table[setitem] = {} end
						other_item_table[setitem][setname] = true
					end
					local fuzzy = idToFuzzy(setitem)
					if not item_table[fuzzy] then item_table[fuzzy] = {} end
					item_table[fuzzy][setname] = true
					if other then
						if not other_item_table[fuzzy] then other_item_table[fuzzy] = {} end
						other_item_table[fuzzy][setname] = true
					end
				end
			end
		end
	end
end

function Baggins_ItemRack:CacheSetsAllCharacters()
	local sets,setname,ids,character

	if not Rack then return end
	if not Rack_User then return end
	if not item_rack_character then return end
	--self:Print("Updating ItemRack sets")

	if other_characters_scanned then
		self:CacheSetsForCharacter(item_rack_character)
	else
		for character in pairs(Rack_User) do
			self:CacheSetsForCharacter(character)
		end
		other_characters_scanned = true
	end

	-- Fix up set table for "Any" character
	local any = {}
	sets_table.Any = any
	for sets in pairs(sets_table[item_rack_character]) do
		any[sets] = true
	end
	for sets in pairs(sets_table.Other) do
		any[sets] = true
	end

	-- Fix up item table for "Any" character
	any = {}
	items_table.Any = any
	for ids,sets in pairs(items_table[item_rack_character]) do
		if not any[ids] then any[ids] = {} end
		for setname in pairs(sets) do
			any[ids][setname] = true
		end
	end
	for ids,sets in pairs(items_table.Other) do
		if not any[ids] then any[ids] = {} end
		for setname in pairs(sets) do
			any[ids][setname] = true
		end
	end
	-- Cause Baggins to update all the bags just in case.
	if sieve then sieve:ForceFullUpdate() end
end

function Baggins_ItemRack:CacheSets()
	if old_version then return self:CacheSetsAllCharacters() end

	local i,set

	local sets = ItemRackUser.Sets

	items_table = {}
	sets_table = {}
	for setname, set in pairs(sets) do
		if not string.match(setname,"^~") then
			sets_table[setname] = true
			for i=0,19 do
				local setitem = set.equip[i]
				if setitem then
					if not items_table[setitem] then items_table[setitem] = {} end
					items_table[setitem][setname] = true
					local fuzzy = idToFuzzy(setitem)
					if not items_table[fuzzy] then items_table[fuzzy] = {} end
					items_table[fuzzy][setname] = true
				end
			end
		end
	end
	if sieve then sieve:ForceFullUpdate() end
end

function Baggins_ItemRack:ItemRack_Sets_Initialize(...)
	self.hooks.ItemRack_Sets_Initialize(...)
	self:CacheSets()
end

function Baggins_ItemRack:ItemRack_Sets_Save_OnClick(...)
	self.hooks.ItemRack_Sets_Save_OnClick(...)
	self:CacheSets()
end

function Baggins_ItemRack:ItemRack_Sets_Remove_OnClick(...)
	self.hooks.ItemRack_Sets_Remove_OnClick(...)
	self:CacheSets()
end

function Baggins_ItemRack:InitCore(...)
	self.hooks[ItemRack].InitCore(...)
	self:CacheSets()
end

function Baggins_ItemRack:ToggleOptions(...)
	self.hooks[ItemRack].ToggleOptions(...)
	self:Unhook(ItemRack,"ToggleOptions")
	self:HookOptions()
end

function Baggins_ItemRack:HookOptions()
	self:Hook(ItemRackOpt,"SaveSet")
	self:Hook(ItemRackOpt,"DeleteSet")
end

function Baggins_ItemRack:SaveSet(...)
	self.hooks[ItemRackOpt].SaveSet(...)
	self:CacheSets()
end

function Baggins_ItemRack:DeleteSet(...)
	self.hooks[ItemRackOpt].DeleteSet(...)
	self:CacheSets()
end
